import logging
from fastapi import APIRouter, HTTPException

from ..noodle import noodle
from ..schemas.dependencies import DependencyRequest

router = APIRouter()
logger = logging.getLogger(__name__)

@router.post('/')
def process_dependencies(req: DependencyRequest):
    try:
        if req.method == 'ADD':
            noodle.add_dependency(req.node_key, req.dependent_node_key)
        elif req.method == 'REMOVE':
            noodle.remove_dependency(req.node_key, req.dependent_node_key)
    
    except Exception as e:
        logger.error(f'Error processing dependencies: {e}')
        raise HTTPException(status_code=500, detail=str(e))


from typing import Literal
from pydantic import BaseModel

class DependencyRequest(BaseModel):
    method: Literal['ADD', 'REMOVE']
    node_key: str
    dependent_key: str
    dependent_url: str | None = None
    
    @property
    def dependent_node_key(self) -> str:
        if self.dependent_url is None:
            return self.dependent_key
        return f'{self.dependent_url}::{self.dependent_key}'


import os
import sys
import time
import json
import requests
import threading
import subprocess
import c_two as cc
from pydantic import BaseModel
from abc import ABCMeta, abstractmethod
from dataclasses import dataclass, field
from typing import TypeVar, Type, Generic, Literal

from .lock import RWLock
from ..scenario import ScenarioNode
from ..schemas.lock import LockInfo
from .server_template import CRM_LAUNCHER_IMPORT_TEMPLATE, CRM_LAUNCHER_RUNNING_TEMPLATE

T = TypeVar('T')

SERVING_TABLE = 'serving'
NODE_KEY = 'node_key'
CONNECTION_COUNT = 'connection_count'

class NodeMessage(BaseModel):
    sender_id: str
    action: str

@dataclass
class SceneNodeRecord:
    node_key: str
    scenario_node: ScenarioNode | None   # None if this is a resource set node, not a resource node
    launch_params: str | None = None
    access_info: str | None = None
    
    parent_key: str | None = None
    children: list['SceneNodeRecord'] = field(default_factory=list)
    
    def add_child(self, child: 'SceneNodeRecord'):
        self.children.append(child)
        self.children.sort(key=lambda child: child.node_key.split('.')[-1].lower())  # sort children by their name
        child.parent_key = self.node_key
    
    def add_children(self, children: list['SceneNodeRecord']):
        for child in children:
            self.add_child(child)
    
    @property
    def is_set(self):
        return self.scenario_node is None

class ISceneNode(Generic[T], metaclass=ABCMeta):
    @property
    @abstractmethod
    def lock(self) -> RWLock:
        raise NotImplementedError

    @property
    @abstractmethod
    def node_key(self) -> str:
        raise NotImplementedError
    
    @property
    @abstractmethod
    def server_scheme(self) -> Literal['local', 'memory', 'http']:
        """Return the scheme of the server address"""
        raise NotImplementedError
    
    @property
    @abstractmethod
    def server_address(self) -> str:
        """Return the server address of the node"""
        raise NotImplementedError
    
    @property
    @abstractmethod
    def crm(self) -> T:
        """Return the CRM instance of the node"""
        raise NotImplementedError
    
    @abstractmethod
    def launch_crm_server(self):
        """Launch the CRM server for the node"""
        raise NotImplementedError
    
    @abstractmethod
    def terminate(self):
        """Terminate the CRM server and release resources"""
        raise NotImplementedError
        
class SceneNode(ISceneNode[T]):
    def __init__(
        self,
        icrm_class: Type[T], record: SceneNodeRecord,   # icrm_class only used for type hinting
        access_mode: Literal['lr', 'lw', 'pr', 'pw'],
        timeout: float | None = None, retry_interval: float = 0.1
    ):
        super().__init__()
        
        access_level = access_mode[0]
        lock_type = access_mode[1]
        
        if lock_type not in ['r', 'w']:
            raise ValueError("lock type must be either 'r' for read or 'w' for write")
        if access_level not in ['l', 'p']:
            raise ValueError("access level must be either 'l' for local or 'p' for process-level")
        
        self._node_key = record.node_key

        self._thread_lock = threading.RLock()

        self._crm: T = None
        self._access_level = access_level
        self._crm_params = record.launch_params
        self._crm_class = record.scenario_node.crm_class
        self._import_script = f'from {record.scenario_node.module} import CRM\n'
        self._lock = RWLock(self._node_key, access_mode, timeout, retry_interval)
    
    @property
    def lock(self) -> RWLock:
        return self._lock
    
    @property
    def node_key(self) -> str:
        return self._node_key
    
    @property
    def server_scheme(self) -> Literal['local', 'memory']:
        if self._access_level == 'l':
            return 'local'
        elif self._access_level == 'p':
            return 'memory'
        else:
            raise ValueError(f'Unknown access mode: {self._access_level}')
    
    @property
    def server_address(self) -> str:
        scheme = ''
        if self._access_level == 'l':
            scheme = 'local://'
        elif self._access_level == 'p':
            scheme = 'memory://'
        return scheme + self._node_key.replace('.', '_') + f'_{self._lock.id}'
    
    def launch_crm_server(self):
        with self._thread_lock:
            scripts = CRM_LAUNCHER_IMPORT_TEMPLATE + self._import_script + CRM_LAUNCHER_RUNNING_TEMPLATE
            
            # Try to launch a CRM server (Process-Level) related to the node
            try:
                # Platform-specific subprocess arguments
                kwargs = {}
                if sys.platform != 'win32':
                    # Unix-specific: create new process group
                    kwargs['preexec_fn'] = os.setsid
                
                cmd = [
                    sys.executable,
                    '-c',
                    scripts,
                    '--server_address', self.server_address,
                    '--node_key', self._node_key,
                    '--params', self._crm_params
                ]
            
                subprocess.Popen(
                    cmd,
                    **kwargs,
                )
            
            except Exception as e:
                raise RuntimeError(f'Failed to launch CRM server for node "{self._node_key}": {e}')
        
    @property
    def crm(self) -> T:
        with self._thread_lock:
            if self._crm is not None:
                return self._crm
            
            self._lock.acquire()

            if self._access_level == 'l':
                params = json.loads(self._crm_params) if self._crm_params else {}
                self._crm = self._crm_class(**params)
                return self._crm
            
            elif self._access_level == 'p':
                self.launch_crm_server()
                
                # Spining up the CRM server, wait for it to be ready
                count = 0
                while cc.rpc.Client.ping(self.server_address, 0.5) is False:
                    if count >= 120: # 60 seconds timeout
                        raise TimeoutError(f'CRM server "{self._node_key}" did not start in time')
                    time.sleep(0.5)
                    count += 1
                
                # Create an ICRM instance related to the node CRM
                self._crm = self._crm_class.__base__()
                
                # Add a C-Two RPC client
                client = cc.rpc.Client(self.server_address)
                self._crm.client = client
                return self._crm
    
    def terminate(self):
        with self._thread_lock:
            if not self._crm is None:
                # For Local-level CRM, terminate it manually
                if self._access_level == 'l':
                    self._crm.terminate()
                
                # For Process-level CRM, just shutdown the C-Two client
                elif self._access_level == 'p':
                    self._crm.client.terminate()
                    cc.rpc.Client.shutdown(self.server_address, -1.0)
                
            # Release the lock
            self._lock.release()

class RemoteSceneNode(ISceneNode[T]):
    def __init__(
        self,
        icrm_class: Type[T], access_info: str,
        access_mode: Literal['lr', 'lw', 'pr', 'pw'],
        timeout: float | None = None, retry_interval: float = 0.1
    ):
        super().__init__()
        
        access_level = access_mode[0]
        lock_type = access_mode[1]
        
        if lock_type not in ['r', 'w']:
            raise ValueError("lock type must be either 'r' for read or 'w' for write")
        if access_level not in ['l', 'p']:
            raise ValueError("access level must be either 'l' for local or 'p' for process-level")
        
        self._node_key = access_info
        self._remote_url, self._remote_key = access_info.split('::')
        
        self._thread_lock = threading.RLock()
        
        self._crm: T = None
        self._icrm_class = icrm_class
        
        self._remote_lock_id: str | None = None
        self._timeout = timeout
        self._lock_type = lock_type
        self._access_level = access_level
        self._retry_interval = retry_interval
    
    @property
    def lock(self) -> RWLock:
        return None  # remote nodes do not have a local lock, they use remote locks
    
    @property
    def node_key(self) -> str:
        return self._node_key
    
    @property
    def server_scheme(self) -> Literal['http']:
        return 'http'
    
    @property
    def server_address(self) -> str:
        # Refer to activate_node(), proxy_node() and deactivate_node() in src/pynoodle/endpoints/proxy.py
        # For more details about the server_address format
        return f'{self._remote_url}/noodle/proxy?node_key={self._remote_key}'
    
    def launch_crm_server(self):
        pass
    
    @property
    def crm(self) -> T:
        with self._thread_lock:
            if self._crm is not None:
                return self._crm
            
            icrm_tag = f'{self._icrm_class.__namespace__}/{self._icrm_class.__name__}/{self._icrm_class.__version__}'
            
            # Get the remote lock from the remote Noodle
            # Refer to activate_node() in src/pynoodle/endpoints/proxy.py for more details about the lock API
            lock_api = (
                f'{self.server_address}&icrm_tag={icrm_tag}&lock_type={self._lock_type}&retry_interval={self._retry_interval}' \
                + (f'&timeout={self._timeout}' if self._timeout is not None else '')
            )
            response = requests.get(lock_api)
            if response.status_code != 200:
                raise RuntimeError(f'Failed to acquire lock for remote CRM server: {response.text}')
            self._remote_lock_id = LockInfo(**response.json()).lock_id
            
            # Create an ICRM instance related to the remote node CRM
            # Refer to proxy_node() in src/pynoodle/endpoints/proxy.py for more details about the proxy API
            proxy_api = f'{self.server_address}&lock_id={self._remote_lock_id}'  # can add &timeout=[-1.0 | <timeout>] if needed
            self._crm = self._icrm_class()
            
            # Add a C-Two RPC client
            client = cc.rpc.Client(proxy_api)
            self._crm.client = client
            return self._crm
    
    def terminate(self):
        with self._thread_lock:
            self._crm.client.terminate()
            
            # Refer to deactivate_node() in src/pynoodle/endpoints/proxy.py for more details about the deactivate API
            deactivate_api = f'{self.server_address}&lock_id={self._remote_lock_id}'
            response = requests.delete(deactivate_api)
            if response.status_code != 200:
                raise RuntimeError(f'Failed to deactivate remote CRM server: {response.text}')

class RemoteSceneNodeProxy(SceneNode[T]):
    def __init__(
        self,
        icrm_class: Type[T], record: SceneNodeRecord,
        access_mode: Literal['lr', 'lw', 'pr', 'pw'],
        timeout: float | None = None, retry_interval: float = 0.1
    ):
        super().__init__(icrm_class, record, access_mode, timeout, retry_interval)
        self._remote_url, self._remote_key = record.access_info.split('::')
        self._remote_lock_id: str | None = None
        self._icrm_class: Type[T] = record.scenario_node.icrm_class

    @property
    def server_scheme(self) -> Literal['http']:
        return 'http'
    
    @property
    def server_address(self) -> str:
        # Refer to activate_node(), proxy_node() and deactivate_node() in src/pynoodle/endpoints/proxy.py
        # For more details about the server_address format
        return f'{self._remote_url}/noodle/proxy?node_key={self._remote_key}'
    
    def launch_crm_server(self):
        pass
    
    @property
    def crm(self) -> T:
        with self._thread_lock:
            if self._crm is not None:
                return self._crm
            
            self._lock.acquire()

            icrm_tag = f'{self._icrm_class.__namespace__}/{self._icrm_class.__name__}/{self._icrm_class.__version__}'

            # Get the twin lock from the remote Noodle
            # Refer to activate_node() in src/pynoodle/endpoints/proxy.py for more details about the lock API
            lock_api = (
                f'{self.server_address}&icrm_tag={icrm_tag}&lock_type={self._lock.lock_type}&retry_interval={self._lock.retry_interval}' \
                + (f'&timeout={self._lock.timeout}' if self._lock.timeout is not None else '')
            )
            response = requests.get(lock_api)
            if response.status_code != 200:
                raise RuntimeError(f'Failed to acquire lock for remote CRM server: {response.text}')
            self._remote_lock_id = LockInfo(**response.json()).lock_id
            
            # Create an ICRM instance related to the remote node CRM
            # Refer to proxy_node() in src/pynoodle/endpoints/proxy.py for more details about the proxy API
            proxy_api = f'{self.server_address}&lock_id={self._remote_lock_id}'  # can add &timeout=[-1.0 | <timeout>] if needed
            self._crm = self._icrm_class()
            
            # Add a C-Two RPC client
            client = cc.rpc.Client(proxy_api)
            self._crm.client = client
            return self._crm
    
    def terminate(self):
        with self._thread_lock:
            self._crm.client.terminate()
            
            # Refer to deactivate_node() in src/pynoodle/endpoints/proxy.py for more details about the deactivate API
            deactivate_api = f'{self.server_address}&lock_id={self._remote_lock_id}'
            response = requests.delete(deactivate_api)
            if response.status_code != 200:
                raise RuntimeError(f'Failed to deactivate remote CRM server: {response.text}')
            self._lock.release()


from .scene import SceneNode
from .noodle import Noodle, noodle
from .c2p import crm, icrm, transferable
from .hooks import NOODLE_INIT, NOODLE_TERMINATE
from .scenario import Scenario, ScenarioNode, RawScenarioNode


CRM_LAUNCHER_IMPORT_TEMPLATE = """
import json
import logging
import argparse
import c_two as cc

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

"""

CRM_LAUNCHER_RUNNING_TEMPLATE = """

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--server_address', type=str, required=True, help='C-Two Server address (e.g., memory://...)')
    parser.add_argument('--node_key', type=str, required=True, help='Node key for the SceneNode')
    parser.add_argument('--params', type=str, help='Json-string of parameters for the CRM')
    args = parser.parse_args()
    
    node_key: str = args.node_key
    server_address: str = args.server_address
    crm_params = json.loads(args.params) if args.params else {}

    crm = CRM(**crm_params)
    server = cc.rpc.Server(server_address, crm, node_key)

    logger.info(f'Starting CRM Server for node {node_key}...')
    server.start()
    
    logger.info(f'CRM Server for node {node_key} started at %s', server_address)
    try:
        server.wait_for_termination()
        server.stop()
    except KeyboardInterrupt:
        logger.info(f'KeyboardInterrupt received, terminating CRM Server for node {node_key}...')
        server.stop()
    finally:
        logger.info(f'CRM Server for node {node_key} terminated.')
"""


import sqlite3
import logging
from typing import Literal

from .config import settings
from .scene.treeger import Treeger, SCENE_TABLE, NODE_KEY

logger = logging.getLogger(__name__)

class Noodle(Treeger):
    def __init__(self):
        super().__init__()
    
    def has_node(self, node_key: str) -> bool:
        """Check if a node exists in the scene database."""
        with sqlite3.connect(settings.SQLITE_PATH) as conn:
            cursor = conn.execute(f'SELECT 1 FROM {SCENE_TABLE} WHERE {NODE_KEY} = ?', (node_key,))
            return cursor.fetchone() is not None
    
    def node_server_address(self, node_key: str, lock_id: str, access_level: Literal['l', 'p']) -> str:
        """Get the server address for a node based on its access level."""
        scheme = ''
        if access_level == 'l':
            scheme = 'local://'
        elif access_level == 'p':
            scheme = 'memory://'
        return scheme + node_key.replace('.', '_') + f'_{lock_id}'

noodle = Noodle()


import shutil
import logging
from fastapi import FastAPI

from .noodle import noodle
from .config import settings
from .endpoints import router
from .scene import RWLock, Treeger

logger = logging.getLogger(__name__)

def NOODLE_INIT(app: FastAPI | None = None) -> None:
        # Initialize Treeger
        Treeger.init()
        
        # Initialize RWLock
        RWLock.init()
            
        # Create a new memory temp directory
        settings.MEMORY_TEMP_PATH.mkdir(parents=True, exist_ok=True)

        # Register Noodle endpoints to the FastAPI app
        if app is not None:
            app.include_router(router, prefix='/noodle', tags=['noodle'])
            for scenario_node in noodle.scenario:
                endpoint = scenario_node.endpoint
                if endpoint is not None:
                    app.include_router(endpoint, prefix='/noodle')

        else:
            logger.debug('No FastAPI app provided, Noodle endpoints will not be registered.')

def NOODLE_TERMINATE() -> None:
    """Terminate Noodle CRM servers running in process level and clean up locks."""
    
    # Shutdown all process-level CRM servers gracefully
    RWLock.release_all_process_servers()
    
    # Forcefully shutdown all nodes' CRM servers running in process level if graceful shutdown fails
    # Nodes's CRM servers running in local level will be automatically shutdown when the process exits
    if settings.MEMORY_TEMP_PATH.exists():
        shutil.rmtree(settings.MEMORY_TEMP_PATH)
        
    # Clear all locks in the database
    RWLock.clear_all()


import json
import socket
import sqlite3
import logging
import requests
from contextlib import contextmanager
from typing import TypeVar, Literal, Type, Generator

from .lock import RWLock
from ..config import settings
from ..schemas.scene import SceneNodeInfo
from ..scenario import scenario_graph, ScenarioNode
from ..schemas.dependencies import DependencyRequest
from .scene_node import ISceneNode, SceneNode, RemoteSceneNode, RemoteSceneNodeProxy, SceneNodeRecord

T = TypeVar('T')
logger = logging.getLogger(__name__)

# DB-related constants
SCENE_TABLE = 'scene'
NODE_KEY = 'node_key'
PARENT_KEY = 'parent_key'
LAUNCH_PARAMS = 'launch_params'
SCENARIO_NODE_NAME = 'scenario_node_name'
ACCESS_INFO = 'access_info' # access URL :: remote node key

DEPENDENCY_TABLE = 'dependency'
DEPENDENT_KEY = 'dependent_key'

class Treeger:
    def __init__(self):
        # Get scenario graph
        self.scenario = scenario_graph

        # Init Noodle access URL
        hostname = socket.gethostname()
        ip = socket.gethostbyname(hostname)
        self.url = f'http://{ip}:{settings.SERVER_PORT}'
    
    @staticmethod
    def init():
        # Create the database file if it doesn't exist
        with sqlite3.connect(settings.SQLITE_PATH) as conn:
            # Create the scene table
            conn.execute(f"""
                CREATE TABLE IF NOT EXISTS {SCENE_TABLE} (
                    {PARENT_KEY} TEXT,
                    {SCENARIO_NODE_NAME} TEXT,
                    {NODE_KEY} TEXT PRIMARY KEY,
                    {ACCESS_INFO} TEXT DEFAULT NULL,
                    {LAUNCH_PARAMS} TEXT DEFAULT NULL,
                    created_at TIMESTAMP DEFAULT CURRENT_TIMESTAMP,
                    FOREIGN KEY ({PARENT_KEY}) REFERENCES {SCENE_TABLE} ({NODE_KEY}) ON DELETE CASCADE
                )
            """)
            conn.execute(f'CREATE INDEX IF NOT EXISTS idx_{PARENT_KEY} ON {SCENE_TABLE}({PARENT_KEY})')
            conn.execute(f'CREATE INDEX IF NOT EXISTS idx_{SCENARIO_NODE_NAME} ON {SCENE_TABLE}({SCENARIO_NODE_NAME})')
            
            # Create the dependency table
            conn.execute(f"""
                CREATE TABLE IF NOT EXISTS {DEPENDENCY_TABLE} (
                    {NODE_KEY} TEXT PRIMARY KEY,
                    {DEPENDENT_KEY} TEXT NOT NULL
                )
            """)
            conn.execute(f'CREATE INDEX IF NOT EXISTS idx_{DEPENDENT_KEY} ON {DEPENDENCY_TABLE}({DEPENDENT_KEY})')
            
            conn.commit()
    
    @contextmanager
    def _connect_db(self):
        """Context manager for database connection."""
        conn = sqlite3.connect(settings.SQLITE_PATH)
        conn.row_factory = sqlite3.Row  # enable column access by name
        conn.execute('PRAGMA foreign_keys = ON;') # enable foreign key support
        try:
            yield conn
        finally:
            conn.close()
    
    def _has_node(self, node_key: str) -> bool:
        """Check if a node exists in the database"""
        with self._connect_db() as conn:
            cursor = conn.execute(f'SELECT 1 FROM {SCENE_TABLE} WHERE {NODE_KEY} = ?', (node_key,))
            return cursor.fetchone() is not None

    def _insert_node(
        self, node_key: str,
        scenario_node_name: str, launch_params: str,
        parent_key: str | None, dependent_node_keys_or_infos: list[str]
    ) -> None:
        """Insert a new node into the database"""
        with self._connect_db() as conn:
            # Add node to the scene table
            conn.execute(f"""
                INSERT INTO {SCENE_TABLE} ({NODE_KEY}, {SCENARIO_NODE_NAME}, {LAUNCH_PARAMS}, {PARENT_KEY})
                VALUES (?, ?, ?, ?)
            """, (
                node_key,
                scenario_node_name,
                launch_params if launch_params else None,
                parent_key if parent_key else None
            ))
            
            # Add dependencies to the dependency table
            for dep_key in dependent_node_keys_or_infos:
                # TODO: This is not atomic operation, if failed, all remote dependency is not added completely
                # Add dependency to the remote Noodle
                if dep_key.startswith('http'): 
                    remote_noodle_url, remote_node_key = dep_key.split('::')
                    req = DependencyRequest(
                        method='ADD',
                        node_key=remote_node_key,
                        dependent_key=node_key,
                        dependent_url=self.url,
                    )
                    response = requests.post(f'{remote_noodle_url}/noodle/dependencies/', json=req.model_dump())
                    if response.status_code != 200:
                        raise requests.RequestException(f'Failed to add dependency {remote_node_key} to remote Noodle {remote_noodle_url}: {response.text}')
                
                conn.execute(f"""
                    INSERT OR IGNORE INTO {DEPENDENCY_TABLE} ({NODE_KEY}, {DEPENDENT_KEY})
                    VALUES (?, ?)
                """, (dep_key, node_key))
            
            conn.commit()

    def _delete_node(self, node_key: str) -> None:
        """Delete a node from the database"""
        with self._connect_db() as conn:
            # Delete node from the scene table
            conn.execute(f'DELETE FROM {SCENE_TABLE} WHERE {NODE_KEY} = ?', (node_key,))
            
            # Find all remote nodes that this node depends on
            cursor = conn.execute(f'SELECT {NODE_KEY} FROM {DEPENDENCY_TABLE} WHERE {DEPENDENT_KEY} = ?', (node_key,))
            keys = [row[NODE_KEY] for row in cursor.fetchall()]
            for key in keys:
                is_remote = key.startswith('http')
                if not is_remote:
                    continue

                # Remove the dependency from the remote Noodle
                server_url, remote_key = key.split('::')
                req = DependencyRequest(
                    method='REMOVE',
                    node_key=remote_key,
                    dependent_key=node_key,
                    dependent_url=self.url,
                )
                response = requests.post(f'{server_url}/noodle/dependencies/', json=req.model_dump())
                if response.status_code != 200:
                    logger.error(f'Failed to remove dependency from remote Noodle: {response.text}')
                    raise requests.RequestException(f'Failed to remove dependency: {response.text}')

            # Delete dependencies from the dependency table
            conn.execute(f'DELETE FROM {DEPENDENCY_TABLE} WHERE {DEPENDENT_KEY} = ?', (node_key,))
            conn.commit()
    
    def _get_child_keys(self, parent_key: str) -> list[str]:
        """Get all child node keys for a given parent from databse"""
        with self._connect_db() as conn:
            cursor = conn.execute(f'SELECT {NODE_KEY} FROM {SCENE_TABLE} WHERE {PARENT_KEY} = ?', (parent_key,))
            return [row[NODE_KEY] for row in cursor.fetchall()]
    
    def _load_node_record(self, node_key: str, is_cascade: bool) -> SceneNodeRecord | None:
        """Load a single node from the database"""
        with self._connect_db() as conn:
            cursor = conn.execute(f"""
                SELECT {NODE_KEY}, {SCENARIO_NODE_NAME}, {LAUNCH_PARAMS}, {PARENT_KEY}, {ACCESS_INFO}
                FROM {SCENE_TABLE}
                WHERE {NODE_KEY} = ?
            """, (node_key,))
            row = cursor.fetchone()
            if row is None:
                return None
            
            # Get SceneNode attributes
            node_key = row[NODE_KEY]
            access_url = row[ACCESS_INFO] if row[ACCESS_INFO] else None
            parent_key = row[PARENT_KEY] if row[PARENT_KEY] else None
            launch_params = row[LAUNCH_PARAMS] if row[LAUNCH_PARAMS] else ''
            scenario_node = self.scenario[row[SCENARIO_NODE_NAME]] if row[SCENARIO_NODE_NAME] else None
            if scenario_node is None:
                logger.error(f'Scenario node {row[SCENARIO_NODE_NAME]} not found in scenario graph')
                return None
            
            # Create SceneNode instance
            node = SceneNodeRecord(
                node_key=node_key,
                parent_key=parent_key,
                access_info=access_url,
                scenario_node=scenario_node,
                launch_params=launch_params
            )
            
            # If not cascade, return the node with no children loaded
            if not is_cascade:
                return node
            
            # Get all children of the node
            children = self._get_child_keys(node_key)
            if children:
                with self._connect_db() as conn:
                    cursor = conn.execute(f"""
                        SELECT {NODE_KEY}, {SCENARIO_NODE_NAME}, {ACCESS_INFO} FROM {SCENE_TABLE} WHERE {PARENT_KEY} = ?
                    """, (node_key,))
                    child_rows = cursor.fetchall()
                    for child_row in child_rows:
                        child_node = SceneNodeRecord(
                            node_key=child_row[NODE_KEY],
                            access_info=child_row[ACCESS_INFO] if child_row[ACCESS_INFO] else None,
                            scenario_node=self.scenario[child_row[SCENARIO_NODE_NAME]] if child_row[SCENARIO_NODE_NAME] else None
                        )
                        node.add_child(child_node)

            return node

    def _is_node_dependency(self, node_key: str) -> bool:
        """Check if a node is depended on by other node"""
        with self._connect_db() as conn:
            cursor = conn.execute(f'SELECT 1 FROM {DEPENDENCY_TABLE} WHERE {NODE_KEY} = ?', (node_key,))
            return cursor.fetchone() is not None
    
    def add_dependency(self, node_key: str, dependent_node_key: str) -> None:
        with self._connect_db() as conn:
            cursor = conn.cursor()
            cursor.execute(f"""
                INSERT INTO {DEPENDENCY_TABLE} ({NODE_KEY}, {DEPENDENT_KEY}) VALUES (?, ?)
            """, (node_key, dependent_node_key))
            conn.commit()
    
    def remove_dependency(self, node_key: str, dependent_node_key: str) -> None:
        with self._connect_db() as conn:
            cursor = conn.cursor()
            cursor.execute(f"""
                DELETE FROM {DEPENDENCY_TABLE} WHERE {NODE_KEY} = ? AND {DEPENDENT_KEY} = ?
            """, (node_key, dependent_node_key))
            conn.commit()
    
    def proxy_node(
        self,
        node_key: str, scenario_node_name: str,
        server_url: str, remote_node_key: str
    ) -> tuple[bool, str]:
        # Check if node already exists in db
        if (self._has_node(node_key)):
            logger.debug(f'Node {node_key} already exists, skipping')
            return True, ''

        try:
            # Add node to the scene
            parent_key = '.'.join(node_key.split('.')[:-1])
            access_info = f'{server_url}::{remote_node_key}'
            with self._connect_db() as conn:
                cursor = conn.cursor()
                cursor.execute(f"""
                    INSERT INTO {SCENE_TABLE} ({NODE_KEY}, {PARENT_KEY}, {SCENARIO_NODE_NAME}, {ACCESS_INFO}) VALUES (?, ?, ?, ?)
                """, (node_key, parent_key, scenario_node_name, access_info))
                conn.commit()
            
            # Add dependency relation to the local Noodle
            self.add_dependency(access_info, node_key)

            # Add dependency relation to the remote Noodle
            req = DependencyRequest(
                method='ADD',
                node_key=remote_node_key,
                dependent_key=node_key,
                dependent_url=self.url,
            )
            response = requests.post(f'{server_url}/noodle/dependencies/', json=req.model_dump())
            if response.status_code != 200:
                logger.error(f'Failed to add dependency to remote Noodle: {response.text}')
                raise requests.RequestException(f'Failed to add dependency: {response.text}')

            logger.info(f'Successfully proxy node "{node_key}" with scenario "{scenario_node_name}"')
            return True, ''
        
        except Exception as e:
            logger.error(f'Failed to proxy node "{node_key}" with scenario "{scenario_node_name}": {e}')
            return False, str(e)

    def mount_node(
        self,
        node_key: str, scenario_node_name: str = '',
        launch_params: any = None, dependent_node_keys_or_infos: list[str] = []
    ) -> tuple[bool, str]:
        # Check if node already exists in db
        if (self._has_node(node_key)):
            logger.debug(f'Node {node_key} already mounted, skipping')
            return True, ''

        # Validate resource SET node
        # If scenario_node_name is not provided
        # Meaning this is a resource set node (no launch params), not a resource node (with launch params or not)
        # If launch_params are provided for a resource set node, raise a warning and ignore launch_params
        if not scenario_node_name and launch_params:
            logger.warning(f'Launch parameters are provided for resource set node "{node_key}", ignoring them')

        scenario_node: ScenarioNode | None = None
        try:
            # Validate resource node
            if scenario_node_name:
                scenario_node = self.scenario[scenario_node_name]
                
                # - Check if the scenario node exists
                if scenario_node is None:
                    raise ValueError(f'Scenario node {scenario_node_name} not found in scenario graph')
                
                # - Check if dependencies are valid
                dep_map: dict[str, bool] = {dep.name: False for dep in scenario_node.dependencies}
                
                # -- Check if all dependencies are provided
                if len(dependent_node_keys_or_infos) != len(dep_map):
                    raise ValueError(f'Node {scenario_node_name} has {len(scenario_node.dependencies)} dependencies, but {len(dependent_node_keys_or_infos)} provided')
                
                # -- Check if all provided dependencies match the scenario node dependencies
                for dep_key in dependent_node_keys_or_infos:
                    # --- Check dependency from a remote node
                    if dep_key.startswith('http'):
                        # ---- Check if it exists in the remote Noodle
                        dep_node_info: SceneNodeInfo | None = None
                        try:
                            # Fetch the node info from the remote Noodle
                            access_url, dep_key = dep_key.split('::')
                            response = requests.get(f'{access_url}/noodle/scene/', params={'node_key': dep_key})
                            
                            # Check if the response is successful
                            if response.status_code != 200:
                                raise ValueError(f'Dependency node {dep_key} not found in remote Noodle {access_url}')
                            
                            # Parse the response to get the node info
                            dep_node_info = SceneNodeInfo.model_validate(response.json())
                            
                        except Exception as e:
                            raise ValueError(f'Failed to fetch dependency node {dep_key} from remote Noodle {access_url}: {e}')
                            
                        # ---- Check if the dependency node has a scenario node name
                        if dep_node_info.scenario_node_name is None:
                            raise ValueError(f'Dependency node {dep_key} in remote Noodle {access_url} is a resource set node, not a resource node')
                        
                        # ---- Check if the dependency node's scenario node exists in the local scenario graph
                        # TODO: This validation is fragile and may not work in all cases
                        #       As it assumes the remote Noodle has the same scenario graph as the local Noodle.
                        #       Consider adding a more robust validation mechanism.
                        if dep_node_info.scenario_node_name not in dep_map:
                            raise ValueError(f'Dependency node {dep_node_info.scenario_node_name} not found in local scenario graph for node {scenario_node_name}')

                        # ---- Mark the dependency as exists
                        dep_map[dep_node_info.scenario_node_name] = True
                        
                    # --- Check dependency from a local node
                    else:
                        # ---- Check if it exists in the scene
                        if not self._has_node(dep_key):
                            raise ValueError(f'Dependency node {dep_key} not found in scene for node {node_key}')
                        
                        # ---- Check if it has a scenario node
                        dep_node_record = self._load_node_record(dep_key, is_cascade=False)
                        if dep_node_record.scenario_node is None:
                            raise ValueError(f'Dependency node {dep_key} is a resource set node, not a resource node')
                        
                        # ---- Mark the dependency as exists
                        dep_map[dep_node_record.scenario_node.name] = True
                    
                if not all(dep_map.values()):
                    missing_deps = [name for name, exists in dep_map.items() if not exists]
                    raise ValueError(f'Missing dependencies for node {scenario_node_name}: {", ".join(missing_deps)}')
            
            # Validate parent key
            parent_key = '.'.join(node_key.split('.')[:-1])
            if parent_key and not self._has_node(parent_key):
                raise ValueError(f'Parent node "{parent_key}" not found in scene for node "{node_key}"')

            # Validate and convert launch parameters
            if scenario_node:
                launch_params = scenario_node.params_converter(node_key, launch_params)

            # If all validations pass, insert node into db
            self._insert_node(node_key, scenario_node_name if scenario_node_name else None, json.dumps(launch_params, indent=4) if launch_params else None, parent_key if parent_key else None, dependent_node_keys_or_infos)
            
            # Call mount hook
            if scenario_node:
                scenario_node.mount(node_key)

            logger.info(f'Successfully mounted node "{node_key}" with scenario "{scenario_node_name}"')
            return True, ''

        except Exception as e:
            logger.error(f'Failed to mount node "{node_key}": {e}')
            return False, str(e)

    def unmount_node(self, node_key: str) -> tuple[bool, str]:
        """Unmount a node from the scene"""
        # If the node does not exist, return True, as it is already unmounted
        if not self._has_node(node_key):
            return True, ''

        try:
            # Try to unmount node recursively
            nodes_count = 0
            node_stack = [node_key]
            nodes_to_delete: list[str] = []
            while node_stack:
                nodes_count += 1
                current_key = node_stack.pop()
                current_node = self._load_node_record(current_key, is_cascade=False)
                
                # If the node is depended, skip it
                if self._is_node_dependency(current_key):
                    continue
                # If the node is active, skip it
                if RWLock.is_node_active(current_key):
                    message = f'Node "{current_key}" is active, cannot unmount node "{node_key}" recursively. Deactivate node "{current_key}" first, then retry unmounting.'
                    raise ValueError(message)
                else:
                    nodes_to_delete.append(current_key)
                
                # If the current node has children (a resource set node)
                # Add all children to the stack for deletion check
                if current_node.is_set:
                    child_keys = self._get_child_keys(current_key)
                    node_stack.extend(child_keys)
            
            if len(nodes_to_delete) != nodes_count:
                message = f'Unable to unmount node "{node_key}": it is depended by {nodes_count - len(nodes_to_delete)} node(s). Remove dependencies first, then retry unmounting.'
                raise ValueError(message)

            # If the node has access information, it is a proxy of a remote node
            # The dependency needs to be removed from the remote Noodle
            access_info = self._load_node_record(node_key, is_cascade=False).access_info
            is_proxy = access_info is not None and '::' in access_info
            if is_proxy:
                server_url, remote_node_key = access_info.split('::')
                req = DependencyRequest(
                    method='REMOVE',
                    node_key=remote_node_key,
                    dependent_key=node_key,
                    dependent_url=self.url,
                )
                response = requests.post(f'{server_url}/noodle/dependencies/', json=req.model_dump())
                if response.status_code != 200:
                    logger.error(f'Failed to remove dependency from remote Noodle: {response.text}')
                    raise requests.RequestException(f'Failed to remove dependency: {response.text}')
            
            # Delete the node from the database
            scenario_node = self._load_node_record(node_key, is_cascade=False).scenario_node
            self._delete_node(node_key)
            
            # Call unmount hook for non-resource-set and non-proxy nodes
            if scenario_node and not is_proxy:
                scenario_node.unmount(node_key)
            
            logger.info(f'Successfully unmounted node "{node_key}"')
            return True, ''

        except Exception as e:
            logger.error(f'Failed to unmount node "{node_key}": {e}')
            return False, str(e)

    def get_node(
        self,
        icrm_class: Type[T] | None, node_key: str,
        access_mode: Literal['lr', 'lw', 'pr', 'pw'],
        timeout: float | None = None, retry_interval: float = 1.0
    ) -> ISceneNode[T] | None:
        # Check if icrm_class is valid (must be an ICRM class)
        if icrm_class and icrm_class.direction == '<-':
            raise ValueError(f'Provided icrm_class {icrm_class.__name__} is a CRM class, provide an ICRM class instead')
        
        # If the node exists in a remote Noodle
        # Return as a RemoteSceneNode
        if node_key.startswith('http'):
            return RemoteSceneNode(
                icrm_class, node_key,
                access_mode, timeout, retry_interval
            )
        
        # Get node record from the scene
        # Check if the node exists and not a resource set
        node_record = self._load_node_record(node_key, is_cascade=False)
        if node_record is None:
            raise ValueError(f'Node "{node_key}" not found in scene tree')
        if node_record.scenario_node is None:
            raise ValueError(f'Node "{node_key}" is a resource set node, cannot get its service')
        
        # If the node is a proxy of a remote node
        # Return as a RemoteSceneNodeProxy
        if node_record.access_info is not None:
            return RemoteSceneNodeProxy(
                icrm_class, node_record,
                access_mode, timeout, retry_interval
            )
        
        # If the node is a local node
        # Return as a SceneNode
        return SceneNode(
            icrm_class, node_record,
            access_mode, timeout, retry_interval
        )
    
    @contextmanager
    def connect_node(
        self,
        icrm_class: Type[T] | None,
        node_key: str,
        access_mode: Literal['lr', 'lw', 'pr', 'pw'],
        timeout: float | None = None,
        retry_interval: float = 1.0
    ) -> Generator[ISceneNode[T], None, None]:
        """Context manager to connect to a node"""
        node = self.get_node(icrm_class, node_key, access_mode, timeout, retry_interval)
        if node is None:
            raise ValueError(f'Node "{node_key}" not found or inaccessible')
        try:
            yield node
        finally:
            node.terminate()
    
    def get_node_info(self, node_key: str, child_start_index: int = 0, child_end_index: int | None = None) -> SceneNodeInfo | None:
        # Check if node exists in the scene
        if not self._has_node(node_key):
            return None
        
        # Load node from the database
        node_record = self._load_node_record(node_key, is_cascade=True)
            
        child_start_index = min(child_start_index, len(node_record.children))
        child_end_index = len(node_record.children) if child_end_index is None else min(child_end_index, len(node_record.children))

        # Get info of child nodes
        children_info: list[SceneNodeInfo] = []
        for child in node_record.children[child_start_index:child_end_index]:
            children_info.append(
                SceneNodeInfo(
                    node_key=child.node_key,
                    access_info=child.access_info,
                    scenario_node_name=child.scenario_node.name if child.scenario_node else None,
                    children=None   # do not focus on children info of children
                )
            )
        
        return SceneNodeInfo(
            node_key=node_record.node_key,
            access_info=node_record.access_info,
            scenario_node_name=node_record.scenario_node.name if node_record.scenario_node else None,
            children=children_info if children_info else None
        )


import asyncio
import logging
import c_two as cc
from typing import Literal
from fastapi import APIRouter, HTTPException, Body, Response

from ..noodle import noodle
from ..scene.lock import RWLock
from ..schemas.lock import LockInfo

router = APIRouter()
logger = logging.getLogger(__name__)

@router.get('/', response_model=LockInfo)
async def activate_node(node_key: str, icrm_tag: str, lock_type: Literal['r', 'w'], timeout: float | None = None, retry_interval: float = 1.0):
    """
    Activates a node in the Noodle system.
    """
    try:
        # Try to get node information
        node_info = noodle.get_node_info(node_key)
        if not node_info:
            raise HTTPException(status_code=404, detail=f'Node {node_key} not found')
        
        # Validate the ICRM tag
        node_icrm_tag = noodle.scenario.get_icrm_tag(node_info.scenario_node_name)
        if node_icrm_tag != icrm_tag:
            raise HTTPException(status_code=404, detail=f'ICRM tag "{icrm_tag}" not match node "{node_key}", expected "{node_icrm_tag}"')

        # Get the node (mock the icrm class not provided)
        node = noodle.get_node(None, node_key, 'p' + lock_type, timeout, retry_interval)

        # Acquire the lock for the node asynchronously
        lock = node._lock
        await lock.async_acquire()
        
        # Launch the node CRM server at process level
        node.launch_crm_server()
        server_address = noodle.node_server_address(node_key, lock.id, 'p')

        # Spin up the CRM server, asynchronously wait for it to be ready
        count = 0
        while cc.rpc.Client.ping(server_address, 0.1) is False:
            if (timeout is not None) and count >= timeout * 10:
                raise TimeoutError(f'CRM server "{node._node_key}" did not start in time')
            await asyncio.sleep(0.1)
            count += 1
        
        return LockInfo(lock_id=lock.id)
    except Exception as e:
        raise HTTPException(status_code=500, detail=f'Error activating node {node_key}: {e}')

@router.post('/')
async def proxy_node(node_key: str, lock_id: str, timeout: float | None = None, body: bytes=Body(..., description='C-Two Event Message in Bytes')):
    """
    Proxies a C-Two event message to the specified node CRM server in process level.
    """
    try:
        # Check if the lock exists
        if not RWLock.has_lock(lock_id):
            raise HTTPException(status_code=404, detail=f'Lock {lock_id} not found for node {node_key}')
        
        server_address = noodle.node_server_address(node_key, lock_id, 'p')
        
        # Relay the message to the node's CRM server asynchronously
        timeout = timeout if timeout is not None else -1.0
        res = await cc.rpc.routing(server_address, body, timeout)
        return Response(res, media_type='application/octet-stream')
    
    except Exception as e:
        raise HTTPException(status_code=500, detail=f'Error proxying node {node_key}: {e}')

@router.delete('/')
def deactivate_node(node_key: str, lock_id: str):
    """
    Deactivates a node in the Noodle system.
    """
    try:
        # Check if the lock exists
        if not RWLock.has_lock(lock_id):
            raise HTTPException(status_code=404, detail=f'Lock {lock_id} not found for node {node_key}')
        
        server_address = noodle.node_server_address(node_key, lock_id, 'p')
        
        # Shutdown the node's CRM server
        cc.rpc.Client.shutdown(server_address, -1.0)
        
        # Remove the local lock
        RWLock.remove_lock(lock_id)
        
    except Exception as e:
        raise HTTPException(status_code=500, detail=f'Error deactivating node {node_key}: {e}')


from pathlib import Path
from pydantic import field_validator
from pydantic_settings import BaseSettings

class Settings(BaseSettings):
    # Server settings
    SERVER_PORT: int = 8000
    
    # Noodle CRM settings
    NOODLE_CONFIG_PATH: str = './noodle.config.yaml'
    
    # Scene database settings
    SQLITE_PATH: Path
    
    # Memory temp directory settings
    MEMORY_TEMP_PATH: Path = Path('./temp')

    @field_validator('SQLITE_PATH', 'MEMORY_TEMP_PATH', mode='before')
    @classmethod
    def validate_path(cls, p: str) -> Path:
        path = Path(p)
        if not path.is_absolute():
            path = Path.cwd() / path
        
        path.resolve()
            
        if not path.exists():
            if path.suffix:
                path.parent.mkdir(parents=True, exist_ok=True)
            else:
                path.mkdir(parents=True, exist_ok=True)
            
        return path

    class Config:
        env_file = '.env'
        extra = 'ignore'
        
settings = Settings()


from pydantic import BaseModel

class ScenarioNodeDescription(BaseModel):
    name: str
    dependencies: list[str] | None = None

class ScenarioConfiguration(BaseModel):
    module_root: str | None = None
    scenario_nodes: list[ScenarioNodeDescription]


from fastapi import APIRouter

from . import scene
from . import proxy
from . import dependencies

router = APIRouter()
router.include_router(dependencies.router, prefix='/dependencies', tags=['noodle/dependencies'])
router.include_router(scene.router, prefix='/scene', tags=['noodle/scene'])
router.include_router(proxy.router, prefix='/proxy', tags=['noodle/proxy'])


from pydantic import BaseModel

class SceneNodeInfo(BaseModel):
    node_key: str
    access_info: str | None = None
    scenario_node_name: str | None = None
    children: list['SceneNodeInfo'] | None = None


import yaml
import logging
import threading
from pathlib import Path
from fastapi import APIRouter
from dataclasses import dataclass
from typing import TypeVar, Type, Callable

from .config import settings
from .schemas.scenario import ScenarioConfiguration

T = TypeVar('T')
logger = logging.getLogger(__name__)

@dataclass
class RawScenarioNode:
    CRM: Type[T] = None
    ICRM: Type[T] = None
    MOUNT: Callable[[str], None] = lambda x: None
    UNMOUNT: Callable[[str], None] = lambda x: None
    PARAM_CONVERTER: Callable[[str, dict | None], dict | None] = lambda x, y: y
    ENDPOINT: APIRouter | None = None
    
    def __post_init__(self):
        if not self.ICRM:
            raise ImportError(f'ICRM class not found in module {self.module}')
        if not self.CRM:
            self.CRM = self.ICRM

@dataclass
class ScenarioNode:
    name: str
    module: str
    dependencies: list['ScenarioNode']
    
    _lock: threading.Lock = threading.Lock()
    
    _crm_class: Type[T] = None
    _icrm_class: Type[T] = None
    _endpoint: APIRouter = None
    _mount: Callable[[str], None] = None
    _unmount: Callable[[str], None] = None
    _params_converter: Callable[[str, dict | None], dict | None] = None
    
    def _load_from_module(self):
        m = __import__(self.module, fromlist=[''])
        raw: RawScenarioNode = getattr(m, 'RAW', None)
        if not raw:
            raise ImportError(f'RawScenarioNode class not found in module {self.module}')
        
        self._mount = raw.MOUNT
        self._crm_class = raw.CRM
        self._icrm_class = raw.ICRM
        self._unmount = raw.UNMOUNT
        self._endpoint = raw.ENDPOINT
        self._params_converter = raw.PARAM_CONVERTER
    
    @property
    def crm_class(self) -> Type[T]:
        with self._lock:
            if self._crm_class is None:
                self._load_from_module()
            return self._crm_class
    
    @property
    def icrm_class(self) -> Type[T]:
        with self._lock:
            if self._icrm_class is None:
                self._load_from_module()
            return self._icrm_class
    
    @property
    def mount(self) -> Callable[[str], None]:
        with self._lock:
            if self._mount is None:
                self._load_from_module()
            return self._mount
    
    @property
    def unmount(self) -> Callable[[str], None]:
        with self._lock:
            if self._unmount is None:
                self._load_from_module()
            return self._unmount
        
    @property
    def params_converter(self) -> Callable[[str, dict | None], dict | None]:
        with self._lock:
            if self._params_converter is None:
                self._load_from_module()
            return self._params_converter

    @property
    def endpoint(self) -> APIRouter | None:
        with self._lock:
            if self._endpoint is None:
                self._load_from_module()
            return self._endpoint
    
    @property
    def icrm_tag(self) -> str:
        cls = self.icrm_class
        
        name = cls.__name__
        version = cls.__version__
        namespace = cls.__namespace__
        
        if not namespace or not name or not version:
            raise ValueError(f'ICRM class {cls.__name__} is missing namespace, name, or version attributes.')
        
        return f'{namespace}/{name}/{version}'

class Scenario:
    def __init__(self):
        # Read configuration
        configuration_path = Path(settings.NOODLE_CONFIG_PATH)
        if not configuration_path.is_absolute():
            configuration_path = Path.cwd() / configuration_path
        with open(configuration_path, 'r') as f:
            config_data = yaml.safe_load(f)
            
        # Parse scenario graph
        config = ScenarioConfiguration(**config_data)
        
        self.graph: dict[str, ScenarioNode] = {}
        
        # - Firstly, create all nodes
        module_root = f'{config.module_root}.' if config.module_root else ''
        for node_description in config.scenario_nodes:
            node = ScenarioNode(
                name=node_description.name,
                module=module_root + node_description.name,
                dependencies=[]
            )
            self.graph[node_description.name] = node

        # - Secondly, resolve dependencies
        for node_description in config.scenario_nodes:
            node = self.graph[node_description.name]
            if node_description.dependencies:
                node.dependencies = [
                    self.graph[dep_name]
                    for dep_name in node_description.dependencies
                ]
    
    def __getitem__(self, scenario_node_name: str) -> ScenarioNode | None:
        if scenario_node_name not in self.graph:
            return None
        return self.graph[scenario_node_name]
    
    def __iter__(self):
        return iter(self.graph.values())

    def get_icrm_tag(self, scenario_node_name: str) -> str | None:
        node = self.graph.get(scenario_node_name)
        if not node:
            return None
        return node.icrm_tag

scenario_graph = Scenario()


import logging
import c_two as cc
from typing import TypeVar

T = TypeVar('T')
logger = logging.getLogger(__name__)

transferable = cc.transferable

def icrm(namespace: str, version: str) -> T:
    if not namespace:
        raise ValueError('Namespace of ICRM cannot be empty.')
    if not version:
        raise ValueError('Version of ICRM cannot be empty (version example: "1.0.0").')
    if not isinstance(version, str) or not version.count('.') == 2:
        raise ValueError('Version must be a string in the format "major.minor.patch".')
    
    def wrapper(cls: T) -> T:
        new_cls = cc.icrm(cls)
        new_cls.__version__ = version
        new_cls.__namespace__ = namespace
        return new_cls
    
    return wrapper

def crm(cls: T) -> T:
    if not hasattr(cls, 'terminate'):
        raise TypeError(f'Class {cls.__name__} does not have a "terminate" method, which is required for CRM functionality.')
    
    return cc.iicrm(cls)


import logging
from fastapi import APIRouter, HTTPException

from ..noodle import noodle
from ..schemas.scene import SceneNodeInfo

router = APIRouter()
logger = logging.getLogger(__name__)

@router.get('/', response_model=SceneNodeInfo)
def get_node_info(node_key: str, child_start_index: int = 0, child_end_index: int = None):
    try:
        node_info = noodle.get_node_info(node_key, child_start_index, child_end_index)
        if not node_info:
            raise HTTPException(status_code=404, detail='Node not found')
        return node_info
    except Exception as e:
        logger.error(f'Error fetching node info: {e}')
        raise HTTPException(status_code=500, detail='Internal Server Error')


import os
import time
import uuid
import asyncio
import logging
import sqlite3
import threading
import c_two as cc
from typing import Literal

from ..config import settings

logger = logging.getLogger(__name__)

class RWLock:
    def __init__(
        self,
        node_key: str,
        access_mode: Literal['lr', 'lw', 'pr', 'pw'],
        timeout: float | None = None,
        retry_interval: float = 1.0
    ):
        if access_mode not in ['lr', 'lw', 'pr', 'pw']:
            raise ValueError("access mode must be either 'lr' for local read, 'lw' for local write, 'pr' for process-level read, or 'pw' for process-level write")

        self.node_key = node_key
        self.access_mode = access_mode
        self.retry_interval = retry_interval
        self.timeout = timeout if (timeout is not None and timeout >= 0) else None
        self.id = f'pid_{os.getpid()}_tid_{threading.get_ident()}_{uuid.uuid4().hex}'
        
    def _get_connection(self):
        """Creates a new database connection."""
        return sqlite3.connect(settings.SQLITE_PATH)
    
    @property
    def access_level(self) -> str:
        return self.access_mode[0]
    
    @property
    def lock_type(self) -> str:
        return self.access_mode[1]
    
    @staticmethod
    def init():
        with sqlite3.connect(settings.SQLITE_PATH) as conn:
            conn.execute("""
                CREATE TABLE IF NOT EXISTS locks (
                    node_key TEXT NOT NULL,
                    lock_type TEXT NOT NULL,
                    lock_id TEXT PRIMARY KEY,
                    access_level TEXT NOT NULL,
                    created_at TIMESTAMP DEFAULT CURRENT_TIMESTAMP
                )
            """)
            conn.commit()
    
    @staticmethod
    def is_node_active(node_key: str) -> bool:
        """Check if a node is currently active."""
        with sqlite3.connect(settings.SQLITE_PATH) as conn:
            cursor = conn.execute('SELECT 1 FROM locks WHERE node_key = ?', (node_key,))
            return cursor.fetchone() is not None
    
    @staticmethod
    def has_lock(lock_id: str) -> bool:
        """Check if a lock with the given ID exists."""
        with sqlite3.connect(settings.SQLITE_PATH) as conn:
            cursor = conn.execute('SELECT 1 FROM locks WHERE lock_id = ?', (lock_id,))
            return cursor.fetchone() is not None
    
    @staticmethod
    def remove_lock(lock_id: str) -> None:
        """Remove a lock with the given ID."""
        with sqlite3.connect(settings.SQLITE_PATH) as conn:
            conn.execute('DELETE FROM locks WHERE lock_id = ?', (lock_id,))
            conn.commit()
    
    @staticmethod
    def release_all_process_servers() -> None:
        """
        Release all process-level locks by shutting down their associated CRM servers.
        
        This function queries the database for all existing locks and specifically handles
        process-level locks (access_level == 'p') by attempting to shutdown their 
        corresponding CRM servers. Local-level locks are ignored in this operation.
        """
        with sqlite3.connect(settings.SQLITE_PATH) as conn:
            cursor = conn.execute('SELECT lock_id, node_key, access_level FROM locks')
            for lock_id, node_key, access_level in cursor.fetchall():
                # For process-level locks, shutdown the CRM server
                if access_level == 'p':
                    server_address = f'memory://{node_key.replace(".", "_")}_{lock_id}'
                    try:
                        if not cc.rpc.Client.shutdown(server_address, -1.0):
                            raise RuntimeError(f'Failed to shutdown CRM server for node {node_key}')
                    except Exception as e:
                        logger.error(f'{e}')
            conn.commit()
    
    @staticmethod
    def clear_all() -> None:
        """Remove all locks from the database."""
        # If no table, do nothing
        if not settings.SQLITE_PATH.exists():
            return

        with sqlite3.connect(settings.SQLITE_PATH) as conn:
            conn.execute('DELETE FROM locks')
            conn.commit()

    def acquire(self) -> None:
        """
        Acquires the lock, blocking until it's available or timeout occurs.
        """
        start_time = time.monotonic()
        while (self.timeout is None) or (time.monotonic() - start_time < self.timeout):
            conn = self._get_connection()
            # Use IMMEDIATE transaction to acquire a reserved lock on the database file,
            # Preventing other connections from writing to the database.
            try:
                conn.execute('BEGIN IMMEDIATE')
                cursor = conn.cursor()
                
                can_acquire = False
                if self.lock_type == 'w':
                    # For a write lock, no other locks should exist for this resource
                    cursor.execute('SELECT COUNT(*) FROM locks WHERE node_key = ?', (self.node_key,))
                    if cursor.fetchone()[0] == 0:
                        can_acquire = True
                else: # 'r'
                    # For a read lock, no write locks should exist for this resource
                    cursor.execute("SELECT COUNT(*) FROM locks WHERE node_key = ? AND lock_type = 'w'", (self.node_key,))
                    if cursor.fetchone()[0] == 0:
                        can_acquire = True
                
                if can_acquire:
                    cursor.execute(
                        'INSERT INTO locks (node_key, lock_type, lock_id, access_level) VALUES (?, ?, ?, ?)',
                        (self.node_key, self.lock_type, self.id, self.access_level)
                    )
                    conn.commit()
                    return
                else:
                    # Could not acquire, rollback and wait.
                    conn.rollback()

            except sqlite3.OperationalError as e:
                # This can happen if another process has an EXCLUSIVE lock (e.g., another BEGIN IMMEDIATE)
                # This is part of the contention mechanism, just rollback and retry.
                conn.rollback()
            finally:
                conn.close()

            # Wait before retrying
            time.sleep(self.retry_interval)
            
        raise TimeoutError(f"Failed to acquire {self.lock_type} lock for resource '{self.resource_name}' within {self.timeout} seconds.")

    def release(self) -> None:
        """Releases the lock."""
        with self._get_connection() as conn:
            try:
                conn.execute('DELETE FROM locks WHERE lock_id = ?', (self.id,))
                conn.commit()
            except Exception as e:
                # Log this error, as failure to release a lock can be critical
                logger.error(f'Error releasing lock {self.id}: {e}')

    async def async_acquire(self) -> None:
        """
        Acquires the lock asynchronously, blocking until it's available or timeout occurs.
        """
        start_time = time.monotonic()
        while (self.timeout is None) or (time.monotonic() - start_time < self.timeout):
            conn = self._get_connection()
            # Use IMMEDIATE transaction to acquire a reserved lock on the database file,
            # Preventing other connections from writing to the database.
            try:
                conn.execute('BEGIN IMMEDIATE')
                cursor = conn.cursor()

                can_acquire = False
                if self.lock_type == 'w':
                    # For a write lock, no other locks should exist for this resource
                    cursor.execute('SELECT COUNT(*) FROM locks WHERE node_key = ?', (self.node_key,))
                    if cursor.fetchone()[0] == 0:
                        can_acquire = True
                else:  # 'r'
                    # For a read lock, no write locks should exist for this resource
                    cursor.execute("SELECT COUNT(*) FROM locks WHERE node_key = ? AND lock_type = 'w'", (self.node_key,))
                    if cursor.fetchone()[0] == 0:
                        can_acquire = True

                if can_acquire:
                    cursor.execute(
                        'INSERT INTO locks (node_key, lock_type, lock_id, access_level) VALUES (?, ?, ?, ?)',
                        (self.node_key, self.lock_type, self.id, self.access_level)
                    )
                    conn.commit()
                    return
                else:
                    # Could not acquire, rollback and wait.
                    conn.rollback()

            except sqlite3.OperationalError as e:
                # This can happen if another process has an EXCLUSIVE lock (e.g., another BEGIN IMMEDIATE)
                # This is part of the contention mechanism, just rollback and retry.
                conn.rollback()
            finally:
                conn.close()

            # Wait before retrying
            await asyncio.sleep(self.retry_interval)

        raise TimeoutError(f"Failed to acquire {self.lock_type} lock for resource '{self.resource_name}' within {self.timeout} seconds.")